@@ -61,16 +61,30 @@ my %os_specific = (
 			'inc' => ['/usr/local/include'],
 			'lib' => ['/usr/local/lib']
 		}
+	},
+	'netbsd' => {
+		'ssh2' => {
+			'inc' => ['/usr/pkg/include'],
+			'lib' => ['/usr/pkg/lib']
+		},
+		'curl' => {
+			'inc' => ['/usr/pkg/include'],
+			'lib' => ['/usr/pkg/lib']
+		}
 	}
 );
 
-my $ssh2_libpath;
-my $ssh2_incpath;
+my ($ssh2_libpath, $ssh2_incpath);
+my ($curl_libpath, $curl_incpath);
 if (my $os_params = $os_specific{$^O}) {
 	if (my $ssh2 = $os_params -> {'ssh2'}) {
 		$ssh2_libpath = $ssh2 -> {'lib'};
 		$ssh2_incpath = $ssh2 -> {'inc'};
 	}
+	if (my $curl = $os_params -> {'curl'}) {
+		$curl_libpath = $curl -> {'lib'};
+		$curl_incpath = $curl -> {'inc'};
+	}
 }
 
 my @library_tests = (
@@ -86,6 +100,8 @@ my @library_tests = (
 	},
 	{
 		'lib'     => 'curl',
+		'libpath' => $curl_libpath,
+		'incpath' => $curl_incpath,
 		'header'  => 'curl/curl.h',
 	},
 );
@@ -161,8 +177,14 @@ $def .= ' -DNO_VIZ -DSTDC -DNO_GZIP -D_FILE_OFFSET_BITS=64 -D_GNU_SOURCE';
 if ($is_solaris) {
 	$def .= ' -D_POSIX_C_SOURCE=200112L -D__EXTENSIONS__ -D_POSIX_PTHREAD_SEMANTICS';
 } elsif ($is_netbsd) {
-	# Needed for stat.st_mtim (atleast on NetBSD)
-	$def .= ' -D_POSIX_C_SOURCE=200809L';
+	# Needed for stat.st_mtim / stat.st_mtimespec
+	$def .= ' -D_NETBSD_SOURCE';
+
+	if ((split (m|\.|, $Config{osvers}))[0] < 7) {
+		$def .= ' -DGIT_USE_STAT_ATIMESPEC';
+	}
+} elsif ($is_osx) {
+	$def .= ' -DGIT_USE_STAT_ATIMESPEC';
 }
 
 if ($is_gcc) {
